"""Vjer CI/CD module.

This module provides utilities for making it easier to write CI/CD automations.

"""

__all__ = ('__title__', '__summary__', '__uri__',
           '__version__', '__build_name__', '__build_date__',
           '__author__', '__email__',
           '__license__', '__copyright__')

__title__ = 'Vjer'
__summary__ = 'CI/CD Toolkit'
__uri__ = 'https://github.com/arisilon/vjer/'

__version__ = '39.0.0rc16'
__build_name__ = 'vjer_39.0.0rc16-14632297672'
__build_date__ = '2025-04-24 02:49:39.002253'

__author__ = 'Jeffery G. Smith'
__email__ = 'web@pobox.com'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2025 Jeffery G. Smith'
